/*
             gps - Gif to color PostScript level 2

                Richard C. Allen, 1993 March 5
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <dos.h>

typedef unsigned char byte;
typedef unsigned long word;
typedef unsigned short sword;

extern double tread();

extern void wrap(  /* the PostScript wrapper generator */
FILE *xpsout,                       /* the output file */
int xsize,                          /* image width */
int ysize,                          /* image height */
double xscale,                      /* image xscale value */
double yscale,                      /* image yscale value */
int ncopies,                        /* number of copies */
int use_lzw,                        /* !0 for LZW compression, else hex */
char *title,                        /* the title string */
void (*xprogress)(int y),           /* progress function */
word (*xreadtrip)(int x,int y));    /* caller's pixel triplet read function */

extern byte *columns[2048];     /* pointers to start of each col */
                                    /* allocated at power up */

extern byte palette[768];       /* original from .gif file */

sysabort(           /* a printf like abort function */
char *fmt,...)
{
    va_list arg_ptr;

    fprintf(stderr,"\n\nabort: ");  /* preamble */
    va_start(arg_ptr, fmt);         /* arg_ptr -> first arg */
    vfprintf(stderr,fmt,arg_ptr);   /* display the message */
    va_end(arg_ptr);                /* arg_ptr = NULL */
    fprintf(stderr,"\n\n");         /* postamble */
    exit(0x7fff);                   /* do the abort */
}

static char *nm[] = {"???","Jan","Feb","Mar","Apr","May","Jun",
                     "Jul","Aug","Sep","Oct","Nov","Dec","???",
                     "???","???"};

/*
**      read epoch from the 'puter
**
*/

static void pc_epoch(
char *s)
{
    struct dos_date_t date;
    struct dos_time_t time;

    dos_getdate(&date);
    dos_gettime(&time);
    sprintf(s,"%d %s %d  %02d:%02d",
        date.year, nm[date.month], date.day,
        time.hour, time.minute);
}

/*
**          make epoch string from diskfile date/time
**
**              yyyy mmm dd  hh:mm:ss
**              0123456789012345678901
**                        111111111122
*/

static void make_sdate(
unsigned short date,    /* in find_t format */
unsigned short time,    /* in find_t format */
char *s)                /* for output */
{
    
    sprintf(s,"%d %s %d  %02d:%02d",
        ((date >> 9) & 0x3f) + 1980,
        nm[(date >> 5) & 0xf],
        date & 0x1f,
        (time >> 11) & 0x1f,
        (time >> 5) & 0x3f);
}

static int file_epoch(      /* returns !0 if ok, else 0 if file not found */
char *path,
char *s)
{
    int fd;
    unsigned date, time;

    fd = dos_open(path,O_RDONLY,0);
    if (fd < 0) return (0);         /* cannot open the pig ... */
    if (dos_getftime(fd,&date,&time))
    {
        sysabort("getftime failed on %s",path);
    }
    close(fd);
    make_sdate(date,time,s);
    return (1);
}

static void progress(int y)           /* progress function */
{
    printf("\routput row %d   ",y);
}

static word readtrip(int x, int y)  /* read rgb value at x,y */
{
    byte *p;
    int z;
    word w;

    p = columns[x];         /* start of column x */
    z = p[y] & 0xff;        /* get the pig ... */
    p = &palette[z*3];      /* -> red value */
    w  = (((int)(*p++)) & 0xff) << 16;
    w |= (((int)(*p++)) & 0xff) << 8;
    w |= (((int)(*p++)) & 0xff);
    return (w);
}


void main(
int ntokens,
char *tokens[])
{
    FILE *ofile;            /* PostScript output file handle */
    short _far *s;          /* -> printer port */
    int ixsize, iysize;     /* input image extents */
    double xscale, yscale;  /* output image scale factors */
    int ncopies;            /* number of copies to print */
    char name[99];          /* filename with no extension */
    char opath[99];         /* output file full pathname */
    char epoch[99];         /* conversion epoch string */
    char gifepoch[99];      /* gif file epoch string */
    char gifname[99];       /* gif file full pathname */
    char program[99];       /* program info */
    char title[199];        /* full title for the print */
    int lzw;                /* !0 for lzw encode, 0 for hex ascii */
    int d;                  /* a Control-D job delimiter */

    lzw = 1;            /* use lzw compression */
    ncopies = 1;        /* default number of copies */

    sprintf(program,
        "GPS - Gif to PostScript - R. C. Allen - version 1.20 1993 Mar 7");

    printf("%s\n",program);             /* announce it */


    if (ntokens < 2)
    {
        sysabort("usage> gps gif_file_name [optional # copies]");
    }
    if (ntokens > 2) ncopies = atoi(tokens[2]);     /* get copy count */
    if (ncopies < 1 || ncopies > 50)    /* check for reasonability */
    {
        sysabort("number of copies < 1 or > 50");
    }

    strcpy(name,tokens[1]);             /* fetch filename head */
    strcpy(gifname,tokens[1]);          /* make .gif */
    strcat(gifname,".gif");             /*  filename ... */

    if (!file_epoch(gifname,gifepoch))  /* find .gif epoch */
    {
        sysabort("cannot open %s",gifname);
    }

    pc_epoch(epoch);                    /* conversion epoch */

    hrsetup();              /* set up the timer */
    start_timer();          /* and start it ... */

    printf("\nreading %s.gif\n",name);      /* announce */
    zrgif(&ixsize,&iysize,64,name);         /* input the .gif file */
    printf("  (%.1f seconds)",tread());     /* show read time */


        /* set up the scaling for common .gif file sizes */
        /* scaling is roughly based on the gif image width */
        /* so that the width will be about 8.5 inches ... */

    if      (ixsize == 1280 && iysize == 1024) xscale = yscale = 2.0;
    else if (ixsize == 1024 && iysize ==  768) xscale = yscale = 2.5;
    else if (ixsize ==  800 && iysize ==  600) xscale = yscale = 3.2;
    else if (ixsize ==  640 && iysize ==  480) xscale = yscale = 4.0;
    else if (ixsize ==  640 && iysize ==  400)
    {
        xscale = 4.0;   /* this is 5/4 aspect ratio */
        yscale = 5.0;
    }
    else if (ixsize == 600 && iysize == 600) xscale = yscale = 3.5;
    else if (ixsize ==  320 && iysize ==  200)
    {
        xscale = 8.0;   /* this is 5/4 aspect ratio */
        yscale = 10.0;
    }
    else 
    {
        printf("\nimage was %d by %d ...",ixsize,iysize);
        printf("\n");
        printf("\nvalid sizes are:");
        printf("\n");
        printf(
  "\n1280x1024, 1024x768, 800x600, 640x480, 640x400, 600x600, and 320x200");
        printf("\n");
        sysabort("invalid image size!");
    }

    start_timer();
    strcpy(opath,name);
    strcat(opath,".ps");
    ofile = fopen(opath,"w");
    if (ofile == NULL) sysabort("cannot open %s for output",opath);
    if (ncopies > 1)
    {
        printf("\noutput to %s %d x %d image scaled %g x %g, %d copies\n",
                        opath,ixsize,iysize,xscale,yscale,ncopies);
    }
    else
    {
        printf("\noutput to %s %d x %d image scaled %g x %g, %d copy\n",
                        opath,ixsize,iysize,xscale,yscale,ncopies);
    }

    sprintf(title,"%s.ps from %s.gif [%s] %d x %d scaled %g x %g [converted on %s by %s]",
                  name, name,gifepoch,  ixsize, iysize, xscale, yscale, epoch, program);

    d = 4;      /* this is the magic control-d */
    if (fwrite(&d,1,1,ofile) < 1) sysabort("write fail");

    wrap(ofile,ixsize,iysize,xscale,yscale,ncopies,lzw,
                title,progress,readtrip);

    d = 4;      /* this is the magic control-d */
    if (fwrite(&d,1,1,ofile) < 1) sysabort("write fail");

    fclose(ofile);
    printf("  (%.1f seconds)\n",tread());

}


