/*
             ztimer - high res timer for zortech

             Richard C. Allen, 1991 November 11

    This function is a replacement for the old hrtime.asm module.

    The start_timer()/tread() routines use the bottom 32 bits
    of the 48 bit epochs and consequently will only work for
    measuring times up to 59.9932 minutes and then will roll over.

    NOTE: these routines do not take care of the midnite rollover!!!!!!
*/

#include <stdio.h>
#include <stdarg.h>
#include <dos.h>
#include <int.h>

typedef unsigned char byte;

static byte _far *low_memory;   /* -> memory starting at zero */
#define LTICK 838.09651e-09     /* length of a tick in seconds */
                                /* 1.0 / 1193180.0 */

typedef unsigned long ulong;    /* type for 32-bit unsigned integers herein */

struct _hrt
{
    ulong slow;             /* the slow ticker count (18.2 Hz) */
    unsigned short fast;    /* the fast ticker value (1 MHz) */
};

static void hdelay(){}  /* time waster */

void hrsetup()      /* this must be called before using the timer */
{
    low_memory = (_x386_mk_protected_ptr)(0);   /* -> low memory */

    int_off();
    outp(0x43,0x34);    /* timer 0, mode 2, 2 bytes, binary */
    hdelay();
    outp(0x40,0);
    hdelay();
    outp(0x40,0);
    int_on();
}

static void hrtime(
struct _hrt *h)
{
    unsigned short low, high;
    ulong _far *bt;

    bt = (ulong _far *) &low_memory[0x46c];     /* -> bios ticker */
    int_off();
    h->slow = *bt;                              /* bios ticker */
    outp(0x43,0);           /* latch the counter */
    hdelay();
    low = inp(0x40);        /* read low byte */
    hdelay();
    high = inp(0x40);       /* read high byte */
    int_on();
    high = (high << 8) | (low & 0xff);
    h->fast = -high;
}

static ulong t0;

ulong qticks()
{
    struct _hrt t;
    ulong tt;

    hrtime(&t);
    tt = (t.slow << 16) | t.fast;
    return (tt);
}

void start_timer()
{
    extern ulong qticks();
    t0 = qticks();    
}

static ulong stop_timer()
{
    extern ulong qticks();
    ulong t1;

    t1 = qticks();
    return (t1-t0);
}

double tread()
{
    extern ulong stop_timer();
    ulong t;
    double tx;

    t = stop_timer();       /* in ticks */
    tx = t;
    tx *= LTICK;            /* in seconds */
    return (tx);
}

#if 0
ulong bticker()
{
    ulong _far *bt;

    bt = (ulong _far *) &low_memory[0x46c];      /* -> bios ticker */
    return(*bt);                                /* bios ticker */
}
#endif

q_delay(
double t)
{
    ulong n, t0, t1;

    if (t < 0.0 || t > 60.0) sysabort("strange argument in q_delay()");

    n = t / LTICK;          /* # 838 ns ticks */
    t0 = qticks();          /* current time */
    t1 = t0 + n;            /* end time */
    for (;;)
    {
        if (qticks() >= t1) return;
    }
}
