/*
            wrap - Zortech version of ps_wrap from Tektronix

                     Richard C. Allen, 1993 March 5

        This code came originally from Rich Knowles at Tektronix.
        It originally input a headerless raw binary rgb file
        and output a PostScript file in either straight hex or
        LZW compressed.  It is "unencumbered" and may be used
        freely although I'm sure the folks at Tektronix would
        like you to buy their printer ...

        I modified it to be called as a function with an external
        routine supplying the image data ...

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#define MSDOSX
#ifdef MSDOSX
#include <fcntl.h>
#include <io.h>
#include <errno.h>
#endif

#define TRUE  (1 == 1)
#define FALSE (1 != 1)
typedef void (* VFUNC_PTR) (short);
typedef unsigned long word;

static FILE *psout;     /* the output file */
extern void sysabort(char *,...);
word (*readtrip)(int x, int y);    /* caller's pixel triplet read function */
void (*progress)(int y);           /* progress function */

static void outchar(int c)
{
    if (fwrite(&c,1,1,psout) < 1)
    {
        sysabort("fwrite failed in wrap:outchar()");
    }
}

/*
 * Public variables required (used by calling routines) **********************
 */

void (* PrintByte)      (short);
void (* LZW_PrintByte)  (short);
void (* RLE_PrintByte)  (short);
void (* CMAP_PrintByte) (short);

struct CMAP {
    short red;
    short green;
    short blue;
} cmap[4096];

/*
 * Defines data for LZW code ************************************************
 */

#define LZW_EOD        257        /* End of Data marker */
#define LZW_CLR        256        /* Clear Table Marker */
#define LZW_MAXSTR  4096    /* using 12 bit max codes */
     
/*
 * typedefs for LZW code ****************************************************
 */

    /*
     * This is the structure for a single entry into the string table
     */
     
typedef struct {
    short prefix;
    short suffix;
    short next;
} stringrec;

/*
 * Static data for LZW code ************************************************
 */

static unsigned long acc       = 0;    /* =bits left over from a sendcode */

static short int     codeWidth = 9;    /* Current code bit width */
static short int     bitsInAcc = 0;    /* Used Bits in acc variable */
static short int     nextIndex;        /*  */
static short int     lastCode  = 0;    /* data value from previous iteration */

static stringrec stringarray[LZW_MAXSTR];  /* This is the string table */

/*
 * Static data for RunLengthEncoding code ***********************************
 */

#define RLE_MINRUN  3    /* 3 appears to give the best results for most files */
#define RLE_MAXLINE (127+RLE_MINRUN)

static unsigned char RLE_color;
static unsigned char RLE_line[RLE_MAXLINE];

static short int     RLE_cnt;

/*
 * Exported entry points from this module **********************************
 */

void CMAP_Unmap      (short);
void CMAP_Unmap4bits (short);
void LZW_Encode      (short);
void RunLengthEncode (short);
void PrintHex        (short);
void PrintASCII85    (short);
void PrintQuotedBin  (short);

/*
 * Internal (not exported) entry points in this module *********************
 */

static void SwitchPipe         (VFUNC_PTR, VFUNC_PTR);
static void LZW_Initialize     (void );
static void LZW_SendCode       (short);
static void LZW_Encode_More    (short);
static int  LZW_FindString     (short *, short);
static void LZW_AddString      (short, short);
static void RunLengthEncodeRep (short);
static void RunLengthEncodeLit (short);
static void RunLengthEncodeEnd (short, VFUNC_PTR);     

/*
 * *************************************************************************
 * GENERAL CODE FOLLOWS
 */

#if 0
usage (argv)
char ** argv;
{
char * progname;

    if (NULL == (progname = strrchr (argv[0], '\\')))
    progname = argv[0];
    else
        progname++;

    fprintf (stderr, "%s: Wrap PostScript code around 24-bit RGB image data.\n", progname);
    fprintf (stderr, "  The resulting file will print on any Level 2 PostScript printer.\n\n");

    fprintf (stderr, "Missing or invalid arguments\n"); 
    fprintf (stderr, "Usage: %s [-abhlr] width height < datafile > psfile\n\n", progname);
/** fprintf (stderr, "Usage: %s [-I[abhlrc]] [-O[abhlrc]]\n\n", progname);
*/
    fprintf (stderr, "where:\n");

/** fprintf (stderr, "  -I -- Input options:\n");
    fprintf (stderr, "      a  -- Decode ASCII85\n");
    fprintf (stderr, "      b  -- Decode Quoted Binary\n");
    fprintf (stderr, "      h  -- Decode ASCII Hex\n");
    fprintf (stderr, "      l  -- Decode LZW\n");
    fprintf (stderr, "      r  -- Decode Run Length\n");
    fprintf (stderr, "      c  -- Expand Color Map to Raw Data\n");
    fprintf (stderr, "  -O -- Output options:\n");
*/
    fprintf (stderr, "     -a  -- ASCII85 Encode\n");
    fprintf (stderr, "     -b  -- Quoted Binary\n");
    fprintf (stderr, "     -h  -- ASCII Hex Encoded\n");
    fprintf (stderr, "     -l  -- LZW Encoded\n");
    fprintf (stderr, "     -r  -- Run Length Encoded\n");
/** fprintf (stderr, "     -c  -- Color Mapped\n");
*/
    fprintf (stderr, "  width  -- width of the image (pixels in one raster)\n");
    fprintf (stderr, "  height -- height of the image (number of rasters)\n");
    fprintf (stderr, "   <     -- standard input redirection from a file, where\n");
    fprintf (stderr, "  datafile is the name of the file containing the bitmap data.\n");
    fprintf (stderr, "   >     -- standard output redirection to a file, where\n");
    fprintf (stderr, "  psfile is the name of the file containing the PostScript data.\n");
    exit (1);
}
#endif

static void do_wrap(
int xsize, int ysize,       /* image extents */
double xscale,              /* image xscale value */
double yscale,              /* image yscale value */
int ncopies,                /* number of copies */
int use_lzw,                /* compression flag */
char *title)                /* the title string */
{
    long int pixel_count = 0;
    int cur_arg = 0;
    int cur_idx = 0;
    int Image_width;
    int Image_height;
    int ch;
    int x, y;
    word w;
    unsigned char in_buf[3];
    
    struct Switches {
        int A85  : 1;
        int QBin : 1;
        int Hex  : 1;
        int LZW  : 1;
        int RLE  : 1;
        int CMap : 1;
    } OutputSwitches;

    Image_width = xsize;
    Image_height = ysize;

    if (use_lzw)    /* gonna do compressed LZW output */
    {
        OutputSwitches.A85  = 1;
        OutputSwitches.QBin = 0;
        OutputSwitches.CMap = 0;
        OutputSwitches.Hex  = 0;
        OutputSwitches.LZW  = 1;
        OutputSwitches.RLE  = 0;
    }
    else    /* gonna do hex output */
    {
        OutputSwitches.A85  = 0;
        OutputSwitches.QBin = 0;
        OutputSwitches.CMap = 0;
        OutputSwitches.Hex  = 1;
        OutputSwitches.LZW  = 0;
        OutputSwitches.RLE  = 0;
    }

    /* Control-D ??? */

    fprintf(psout,"%%!\n");
    
    fprintf(psout,"%%  print the title\n");
    fprintf(psout,"/Helvetica findfont 6 scalefont setfont\n");
    fprintf(psout,"gsave\n");
    fprintf(psout,"90 rotate\n");
    fprintf(psout,"0 -8 72 mul translate\n");
    fprintf(psout,"144 5 moveto\n");
    fprintf(psout,"(%s) show\n",title);
    fprintf(psout,"grestore\n");
    fprintf(psout,"\n");

    fprintf(psout,"/pixels  %d def\n", Image_width);
    fprintf(psout,"/rasters %d def\n", Image_height);
    fprintf(psout,"/infile currentfile ");

    if (OutputSwitches.A85)
    {
        PrintByte = PrintASCII85;
        fprintf(psout,"/ASCII85Decode filter ");
    }
    else
    if (OutputSwitches.QBin)
    {
        PrintByte = PrintQuotedBin;
        fprintf(psout,"");
    }
    else
    if (OutputSwitches.Hex)
    {
        PrintByte = PrintHex;
        fprintf(psout,"/ASCIIHexDecode filter ");
    }

    if (OutputSwitches.LZW)
    {
    LZW_PrintByte = PrintByte;
        PrintByte     = LZW_Encode;
    fprintf(psout,"/LZWDecode filter ");
    }

    if (OutputSwitches.RLE)
    {
    RLE_PrintByte = PrintByte;
        PrintByte     = RunLengthEncode;
    fprintf(psout,"/RunLengthDecode filter ");
    }

    fprintf(psout,"def\n");

    if (OutputSwitches.CMap)
    {
    CMAP_PrintByte = PrintByte;
        PrintByte      = CMAP_Unmap;
    fprintf(psout,"[ /Indexed /DeviceRGB %d %%%% Insert ColorMap size here\n", 0);
    fprintf(psout,"< FFFFFF 000000 00FF00 .... %%%% Insert Hex ColorMap Data here");
    fprintf(psout,">\n] setcolorspace\n");
    }
    else
    {
    fprintf(psout,"/DeviceRGB setcolorspace\n");
    }

    fprintf(psout,"\n%%%% Center the image in the page and scale it one-to-one\n");
    fprintf(psout,"90 rotate\n");
    fprintf(psout,"clippath pathbbox\n");
    fprintf(psout,"exch 4 -1 roll add 2 div 3 1 roll add 2 div translate\n");
    fprintf(psout,"pixels 300 72 div div rasters 300 72 div div scale\n");
    fprintf(psout,"%g %g scale\n\n",xscale,yscale);

    fprintf(psout,"<< /ImageType 1\n");
    fprintf(psout,"   /Width  pixels\n");
    fprintf(psout,"   /Height rasters\n");
    fprintf(psout,"   /BitsPerComponent %d\n", 8);
    fprintf(psout,"   /ImageMatrix [pixels 0 0 rasters neg pixels 2 div rasters 2 div]\n");
    fprintf(psout,"   /Interpolate false\n");
    if (OutputSwitches.CMap)
        fprintf(psout,"   /Decode [0 %d]  %%%% Insert ColorMap size minus 1 here\n", 255);
    else
        fprintf(psout,"   /Decode [0 1 0 1 0 1]\n");
    fprintf(psout,"   /DataSource infile\n");
    fprintf(psout,">> image\n");

#if 0
#ifdef MSDOSX
    (void) setmode (fileno(stdin), O_BINARY);
#endif
#endif

#if 0
    if (OutputSwitches.CMap)
    {
        while (EOF != (ch = getchar()))
        {
            PrintByte (ch);
            ++pixel_count;
        }
    }
    else
    {
        while (read (fileno(stdin), in_buf, 3) == 3)
        {
            PrintByte (in_buf[0]);
            PrintByte (in_buf[1]);
            PrintByte (in_buf[2]);
            ++pixel_count;
        }
    }
    
#endif

    for (y = 0; y < ysize; y++)
    {
        (*progress)(y);
        for (x = 0; x < xsize; x++)
        {
            w = (*readtrip)(x,y);
            PrintByte ((w >> 16) & 0xff);       /* red */
            PrintByte ((w >> 8) & 0xff);        /* green */
            PrintByte (w & 0xff);               /* blue */
        }
    }

    PrintByte ((short) -1);

    if (ncopies > 1) fprintf(psout,"\n/#copies %d def\n",ncopies);
    fprintf(psout,"\nshowpage\n");

    /* Control-D ??? */

#if 0
    if (pixel_count < ((long) Image_width * Image_height))
    {
        fprintf(stderr, "Out of Data before all pixel data read\n");
    fprintf(stderr, "Data read: %ld,  Data wanted %ld\n",
        pixel_count, ((long) Image_width * Image_height));
    }
    else
    if (pixel_count > ((long) Image_width * Image_height))
    {
        fprintf(stderr, "Too much pixel data read\n");
    fprintf(stderr, "Data read: %ld,  Data wanted %ld\n",
        pixel_count, ((long) Image_width * Image_height));
    }
#endif

#if 0
    return(0);
#endif
}

/*
 * *************************************************************************
 * SUPPORT CODE FOLLOWS
 */

static void SwitchPipe (old_handler, new_handler)
  VFUNC_PTR old_handler;
  VFUNC_PTR new_handler;
{
  if ((VFUNC_PTR) PrintByte      == old_handler)
      (VFUNC_PTR) PrintByte       = new_handler;
  else
  if ((VFUNC_PTR) LZW_PrintByte  == old_handler)
      (VFUNC_PTR) LZW_PrintByte   = new_handler;
  else
  if ((VFUNC_PTR) RLE_PrintByte  == old_handler)
      (VFUNC_PTR) RLE_PrintByte   = new_handler;
  else
  if ((VFUNC_PTR) CMAP_PrintByte == old_handler)
      (VFUNC_PTR) CMAP_PrintByte  = new_handler;

  return;
}


/*
 * *************************************************************************
 * ColorMap related CODE FOLLOWS
 */

void CMAP_Unmap (short pixel)
{
  if (pixel >= 0)
  {
    (*CMAP_PrintByte) (cmap[pixel].red);
    (*CMAP_PrintByte) (cmap[pixel].green);
    (*CMAP_PrintByte) (cmap[pixel].blue);
  }
  else
  {
    (*CMAP_PrintByte) (pixel);
  }
  return;
}

void CMAP_Unmap4bits (short pixel)
{
  if (pixel >= 0)
  {
    (*CMAP_PrintByte) (cmap[pixel >> 4].red);
    (*CMAP_PrintByte) (cmap[pixel >> 4].green);
    (*CMAP_PrintByte) (cmap[pixel >> 4].blue);

    (*CMAP_PrintByte) (cmap[pixel & 0xf].red);
    (*CMAP_PrintByte) (cmap[pixel & 0xf].green);
    (*CMAP_PrintByte) (cmap[pixel & 0xf].blue);
  }
  else
  {
    (*CMAP_PrintByte) (pixel);
  }
  return;
}

/*
 * *************************************************************************
 * LZW related CODE FOLLOWS
 */

void LZW_Encode (short value)
{
  LZW_SendCode   (LZW_CLR);
  LZW_Initialize ();

  if (value >= 0)
  {
    lastCode = value;

    SwitchPipe (LZW_Encode, LZW_Encode_More);
  }
  else
  {
    LZW_SendCode  (LZW_EOD);

    if (bitsInAcc != 0)  /* flush out last byte, if necessary */
    {
      LZW_PrintByte ((short) (acc >> 24));
    }

    LZW_PrintByte (value);
  }
}

static void LZW_Initialize ()
{
int index;

  for (index = 0; index < 256; index++)
  {
    stringarray[index].prefix = -1;
    stringarray[index].suffix = (char) index;
    stringarray[index].next   = -1;
  }
  nextIndex = LZW_EOD + 1;
  codeWidth = 9;
}

static void LZW_SendCode (short theCode)
{
  acc       += (long) theCode << (32 - codeWidth - bitsInAcc);
  bitsInAcc += codeWidth;

  while (bitsInAcc >= 8)
  {
    LZW_PrintByte ((short) (acc >> 24));
    acc = acc << 8;
    bitsInAcc -= 8;
  }

  return;
}

static void LZW_Encode_More (short value)
{
  if (value == -1)
  {
    LZW_SendCode  (lastCode);
    LZW_SendCode  (LZW_EOD);

    if (bitsInAcc != 0)  /* flush out last byte, if necessary */
    {
      LZW_PrintByte ((short) (acc >> 24));
    }

    LZW_PrintByte (value);

    /*
     * Reinitialize for next set of data (if any)
     */

    acc       = 0;
    codeWidth = 9;
    bitsInAcc = 0;

    SwitchPipe (LZW_Encode_More, LZW_Encode);
  }
  else
  {
    if ( ! LZW_FindString (&lastCode, value))
    {
      LZW_SendCode  (lastCode);
      LZW_AddString (lastCode, value);
      lastCode = value;
    }
  }

  return;
}

static int LZW_FindString (short *prefix, short suffix)
{
short index = *prefix;

  while (index != -1)
  {
    if (   (stringarray[index].prefix == *prefix)
        && (stringarray[index].suffix ==  suffix))
    {
      *prefix = index;
      return(TRUE);
    }
    else
    {
      index = stringarray[index].next;
    }
  }

  return(FALSE);
}

static void LZW_AddString (short prefix, short suffix)
{

  stringarray[nextIndex].prefix = prefix;
  stringarray[nextIndex].suffix = suffix;
  stringarray[nextIndex].next   = stringarray[prefix].next;
  stringarray[prefix].next      = nextIndex;

  if ((++nextIndex >> codeWidth) != 0) /* Did we just move up to next bit width? */
  {
    if (++codeWidth > 12) /* Did we overflow the max bit width? */
    {
      --codeWidth;
      LZW_SendCode   (LZW_CLR);
      LZW_Initialize ();
    }
  }

  return;
}



/*
 * *************************************************************************
 * RunLengthEncoding related CODE FOLLOWS
 */

void RunLengthEncode (short value)
{
  if (value >= 0)
  {
    RLE_cnt     = 1;
    RLE_color   = value;
    RLE_line[0] = value;
    RLE_line[1] = value;
    SwitchPipe (RunLengthEncode, RunLengthEncodeRep);
  }
  else
  {
    (*RLE_PrintByte) (0x80); /* RLE EndOfData Marker */
    (*RLE_PrintByte) (value);
  }
}

static void RunLengthEncodeRep (short value)
{
  if (value >= 0)
  {
  unsigned char color = value;
    /*
     * If the colors are the same, just keep counting, but make sure that the
     * run length will fit in a signed byte (127 max, i.e.  255 means length
     * 2, 254 means length 3, 129 means length 128 etc:  count = 257 - length).
     * Lengths of 0 to 127 indicate that 1 to 128 literal bytes follow.
     */

    if (color == RLE_color)
    {
      if (RLE_cnt == 128) /* check for max run length */
      {
    (*RLE_PrintByte) (257 - RLE_cnt);
    (*RLE_PrintByte) (RLE_color);
    RLE_cnt = 0;
      }
      RLE_cnt++;
    }
    else
    {
      if (RLE_cnt >= RLE_MINRUN)
      {
    /*
     * The color changed with a worthwhile run, so output the last run and
     * prepare for another run to start.
     */

    (*RLE_PrintByte) (257-RLE_cnt);
    (*RLE_PrintByte) (RLE_color);

    RLE_cnt     = 1;
    RLE_color   = color;
    RLE_line[0] = color;
    RLE_line[1] = color;
      }
      else
      {
    /*
     * The color changed with only a run of one or two, so form a 'literal
     * run' of characters, stopping when the next series of bytes in the
     * raster are identical (and thus will form a 'repeat run').
     */

    RLE_line[RLE_cnt++] = color;
    SwitchPipe (RunLengthEncodeRep, RunLengthEncodeLit);
      }
    }
  }
  else
  {
    RunLengthEncodeEnd (value, RunLengthEncodeRep);
  }
}

static void RunLengthEncodeLit (short value)
{
int k;

  if (value >= 0)
  {
  unsigned char color = value;

    RLE_line[RLE_cnt++] = color;

    if (RLE_cnt > RLE_MINRUN)
    {
      if (   (RLE_line[RLE_cnt - (RLE_MINRUN - 0)] == color)
#if RLE_MINRUN > 2
      && (RLE_line[RLE_cnt - (RLE_MINRUN - 1)] == color)
#if RLE_MINRUN > 3
      && (RLE_line[RLE_cnt - (RLE_MINRUN - 2)] == color)
#if RLE_MINRUN > 4
      && (RLE_line[RLE_cnt - (RLE_MINRUN - 3)] == color)
#endif
#endif
#endif
         )
      {
    /*
     * Found a run so output the literal run that preceeds that repeat run
     * and return to repeat run mode.
     */

    RLE_cnt--;
    (*RLE_PrintByte) (RLE_cnt - RLE_MINRUN);
    for (k = 0; k <= (RLE_cnt - RLE_MINRUN); k++)
    {
      (*RLE_PrintByte) (RLE_line[k]);
    }
    RLE_cnt     = 3;
    RLE_color   = color;
    SwitchPipe (RunLengthEncodeLit, RunLengthEncodeRep);
      }
      else
      if (RLE_cnt >= RLE_MAXLINE)
      {
    /*
     * Max literal run of 128 reached so output that data and realign
     * structures for more.  (Actually, a RLE_line holds a little more
     * than 128 entries so that the 127 and 128th entries can participate
     * in repeat pattern checking with next couple of bytes.)
     */

    (*RLE_PrintByte) (127);
    for (k = 0; k <= (127); k++)
    {
      (*RLE_PrintByte) (RLE_line[k]);
    }

    for (k = 128; k < RLE_MAXLINE; k++)
    {
      RLE_line[k-128] = RLE_line[k];
    }
    RLE_cnt = k - 128;
      }
    }
  }
  else
  {
    RLE_cnt--;
    (*RLE_PrintByte) (RLE_cnt);
    for (k = 0; k <= (RLE_cnt); k++)
    {
      (*RLE_PrintByte) (RLE_line[k]);
    }

    RLE_cnt = 0;

    RunLengthEncodeEnd (value, RunLengthEncodeLit);
  }
}

static void RunLengthEncodeEnd (short value, VFUNC_PTR last_handler)
{
  SwitchPipe (last_handler, RunLengthEncode);

  if (RLE_cnt > 0)
  {
    /*
     * Flush out any remaining run data.  Mask of 0x0ff used to cover the case
     * where RLE_cnt == 1 and thus a zero (0) result is desired.
     */

    (*RLE_PrintByte) ((257 - RLE_cnt) & 0x0ff);
    (*RLE_PrintByte) (RLE_color);
  }

  (*RLE_PrintByte) (0x80);  /* write RunLength marker */
  (*RLE_PrintByte) (value); /* write EOD marker */

}



/*
 * *************************************************************************
 * HEX/ASCII85/Binary related CODE FOLLOWS
 */

/*
 * PrintByte() - Actually a pointer to one of several output formats.
 * Encode the output either as Quoted Binary, Hex, or ASCII85 bytes.  Quoted
 * Binary adds one character for each occurance of one of 8 control
 * characters.  Hex encoding results in 2 bytes output for every one input.
 * ASCII85 results in 5 bytes output for every 4 bytes input.  An input value
 * of -1 is treated as an EOF indicator and causes appropriate EndOfData
 * indicators to be written.  Input data values are expected to be in the
 * range of 0 to 255 (and forced there if not!).
 */

void PrintHex (short value)
{
static short outputcount = 0;

    if (value == -1) /* EOF indication */
    {
    outchar ('>'); /* Hex EOD Marker */
        outchar ('\n');
    }
    else
    {
    fprintf(psout,"%02x", (value & 0x0ff));
    outputcount++;
    /*
     * Check to make sure the output lines are not too long
     */
    if (outputcount > 32)
    {
      outchar ('\n');
      outputcount = 0;
    }
    }
}

void PrintASCII85 (short value)
{
static long          outputcount = 0;
static int           valuecount  = 0;
static unsigned long outputvalue = 0;

    if (value == -1) /* EOF indication */
    {
      if (valuecount > 0)
      {
    outputvalue <<= (8 * (4 - valuecount));

    outchar ((int) ((outputvalue / ((long) 85 * 85 * 85 * 85)) + '!'));
    outputvalue %= ((long) 85 * 85 * 85 * 85);

    outchar ((int) ((outputvalue / ((long) 85 * 85 * 85)) + '!'));
    outputvalue %= ((long) 85 * 85 * 85);

    if (valuecount > 1)
    {
      outchar ((int) ((outputvalue / ((long) 85 * 85)) + '!'));
      outputvalue %= ((long) 85 * 85);
    }

    if (valuecount > 2)
    {
      outchar ((int) ((outputvalue / ((long) 85)) + '!'));
      outputvalue %= ((long) 85);
    }

    outputcount = 1;  /* any non-zero value will do */
      }

      if (outputcount != 0)
          outchar ('\n');

      outchar ('~');  /* ASCII85 EOD ... */
      outchar ('>');  /* ... marker      */
      outchar ('\n');

      outputcount = 0;
      valuecount  = 0;
      outputvalue = 0;
    }
    else
    {
      outputvalue <<= 8;
      outputvalue |= (long) (value & 0x0ff);
      if (++valuecount >= 4)
      {
    valuecount = 0;
    if (outputvalue == 0) /* Special case for all zeroes */
    {
      outchar ('z');
      outputcount++;
    }
    else
    {
      outchar ((int) ((outputvalue / ((long) 85 * 85 * 85 * 85)) + '!'));
      outputvalue %= ((long) 85 * 85 * 85 * 85);
      outchar ((int) ((outputvalue / ((long) 85 * 85 * 85)) + '!'));
      outputvalue %= ((long) 85 * 85 * 85);
      outchar ((int) ((outputvalue / ((long) 85 * 85)) + '!'));
      outputvalue %= ((long) 85 * 85);
      outchar ((int) ((outputvalue / ((long) 85)) + '!'));
      outputvalue %= ((long) 85);
      outchar ((int) ((outputvalue) + '!'));
      outputcount += 5;
    }
    /*
     * Check to make sure the output lines are not too long (since not hex
     * we allow twice the data amount on the line)
     */
    if (outputcount/2 > 35)
    {
      outchar ('\n');
      outputcount = 0;
    }
      }
    }
}

void PrintQuotedBin (short value)
{
static short quoteflags[] =
  {
     FALSE,    /* 0x00 */
     TRUE,    /* 0x01        Control-A = Quote Char */
     FALSE,    /* 0x02 */
     TRUE,    /* 0x03        Control-C = Abort & Flush */
     TRUE,    /* 0x04        Control-D = End of Job */
     TRUE,    /* 0x05        Control-E = (reserved for future) */
     FALSE,    /* 0x06 */
     FALSE,    /* 0x07 */
     FALSE,    /* 0x08 */
     FALSE,    /* 0x09 */
     FALSE,    /* 0x0A */
     FALSE,    /* 0x0B */
     FALSE,    /* 0x0C */
     FALSE,    /* 0x0D */
     FALSE,    /* 0x0E */
     FALSE,    /* 0x0F */
     FALSE,    /* 0x10 */
     TRUE,    /* 0x11        Control-Q = XON char */
      FALSE,    /* 0x12 */
     TRUE,    /* 0x13        Control-S = XOFF char */
     TRUE,    /* 0x14        Control-T = Status Request */
      FALSE,    /* 0x15 */
      FALSE,    /* 0x16 */
      FALSE,    /* 0x17 */
      FALSE,    /* 0x18 */
      FALSE,    /* 0x19 */
      FALSE,    /* 0x1A */
      FALSE,    /* 0x1B */
      TRUE,    /* 0x1C        Control-\ = (reserved for future) */
  };

    if (value == -1) /* EOF indication */
    {
      outchar ('\n');
      /* Do nothing */
    }
    else
    {
      if ((value <= 0x1C) && (quoteflags[value]))
      {
    outchar ('A' & ~0x40); /* ctrl-A, the quote charater */
    value ^= 0x40; /* non-ctrl the value */
      }
      outchar (value);

      /*
       * Do not insert newlines since they will be read as data!!
       */
    }
}

/*
        wrap will wrap a PostScript envelope around caller supplied
        data, writing the output to the caller supplied file handle.

        wrap will make xsize*ysize calls to readtrip() which will
        return an rgb triplet in a 32-bit word.

                3322222222221111111111
                10987654321098765432109876543210
                --------rrrrrrrrggggggggbbbbbbbb

*/

void wrap(
FILE *xpsout,               /* the output file */
int xsize,                  /* image width */
int ysize,                  /* image height */
double xscale,              /* image xscale value */
double yscale,              /* image yscale value */
int ncopies,                /* number of copies */
int use_lzw,                /* !0 for LZW compression, else hex */
char *title,                /* the title string */
void (*xprogress)(int y),           /* progress function */
word (*xreadtrip)(int x,int y))     /* caller's pixel triplet read function */
{
    psout = xpsout;         /* save the FILE * */
    progress = xprogress;   /*  and the caller's progress function */
    readtrip = xreadtrip;   /*  and the caller's pixel read function */

    do_wrap(xsize,ysize,xscale,yscale,ncopies,use_lzw,title);
}

