/*
                    zrgif - read .gif files

                 Richard C. Allen, 1993 March 3

    zrgif() will read the gif file into memory and return the
    image size.  set_grey_scale(int nz) will compute a palette
    of values in 0,nz for the image.  read_gif_pixel(int x, int y)
    is then called for pixel values.  when done with the image,
    close_gif() will free the allocated image memory ...

    For the Tektronix Phaser IIsd version, the image is referenced
    directly from the 'columns' and 'palette' arrays.  Grey-scale
    versions may use read_gif_pixel() which reference 'hpalette'.

    GIF and 'Graphics Interchange Format' are trademarks (tm) of Compuserve
    Incorporated, an H&R Block Company.

    The original decoding algorithm came from a public domain C program
    call 85gif.c by Kyle Powell (76044,2215) and was ported to protected
    mode code.

*/


#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <dos.h>
#include <ctype.h>

typedef unsigned int word;
typedef unsigned char byte;

extern void sysabort(char *,...);

#define MAXNCOLS 2048       /* maximum # columns in image */

static short ncols;   
static short nrows;

byte *columns[MAXNCOLS];     /* pointers to start of each col */
                                    /* allocated at power up */

byte palette [768];     /* original from .gif file */
byte hpalette [256];    /* for halftone */

#if 0
static byte outrow[1024] = {0};
#endif
static byte outrow[MAXNCOLS] = {0};
static byte ostack[4096] = {0};

static byte cmask[9] = {0,1,3,7,0xf,0x1f,0x3f,0x7f,0xff};

static unsigned short rem = 0;

static short remct = 0;

static FILE *inf;           /* the .gif file */
static short bufct;           /* # bytes left in the current block */

static short xadj, yadj;      /* offset for display centering */

static short x,y;             /* current location in image */
static short ix, iy;          /* screen location for 0,0 of image */
static short iw, ih;          /* image width and height */

static short pass;            /* interlace pass counter */

static short rowcnt;          /* # pixels remaining in current row */
static short done;            /* ?? */
static short interlaced;      /* !0 if interlaced mode being processed */

static unsigned char ctfirst[4096] = {0};
static unsigned char ctlast[4096] = {0};
static short ctlink[4096] = {0};

static short nextcode = 0;
static short nextlim = 0;


/*
**        initialise the screen at start-up
**
*/

static void initscreen(height, width)
int height, width;
{
}

/*
**        close the screen at shutdown
**
*/

static void closescreen()
{
}

static short nextbyte(void)
{
    short d;

    if (fread(&d,1,1,inf) < 1)
    {
        sysabort("premature eof in nextbyte()");
    }
    return (d & 0xff);
}
/*
**        flush the current image
**
*/

static void flushin()
{
    for (;;)
    {
        while (bufct)       /* 'til current block empty */
        {
            nextbyte();
            bufct--;
        }
        bufct = nextbyte();    /* 'til no more blocks */
        if (bufct == 0) break;
    }
}


/*
**        get next byte from the block
**
*/

static unsigned char getgb()
{
    unsigned char gb;

    if (bufct == 0)     /* no more in current block */
    {
        bufct = nextbyte();      /* get next block byte count */

        if ((bufct == EOF) || (bufct == 0))     /* premature eof */
        {
            closescreen();
            printf("Premature end of data");
            exit(2);
        }
    }
    if ((gb = nextbyte()) == EOF)    /* get next byte from block */
    {
        closescreen();
        printf("PrematureEOF");
        exit(1);
    }
    bufct--;            /* detally the byte */
    return(gb);         /* and return it */
}

/*
**        
**
*/

static unsigned char getbcode(reqct)
int    reqct;
{
    unsigned char retcode;

    if (remct == 0) {
        rem = getgb();
        remct = 8;
    }
    if (remct < reqct) {
        rem = rem | (((int) getgb()) << remct);
        remct += 8;
    }
    retcode = rem & cmask[reqct];
    remct -= reqct;
    rem = rem >> reqct;
    return(retcode);
}


static unsigned short getcode(reqct)
int reqct;
{
    short    temp;

    if (reqct <= 8) return(getbcode(reqct));
    temp = getbcode(8);
    return((getbcode(reqct-8) << 8) | temp);
}


static void init_codetable(cc)
int    cc;
{
    short    code = 0;

    nextcode = cc + 2;
    nextlim = cc << 1;
    while (code < cc) {
        ctfirst[code] = ctlast[code] = code;
        ctlink[code++] = -1;
    }
    while (code < 4096) ctlink[code++] = -2;
}


static void insert_code(code, oldcode, csizeptr)
short code, oldcode, *csizeptr;
{
    ctlink[nextcode] = oldcode;
    ctlast[nextcode] = ctfirst[code];
    ctfirst[nextcode] = ctfirst[oldcode];
    if (++nextcode == nextlim) {
        if (*csizeptr < 12) {
            *csizeptr += 1;
            nextlim <<= 1;
        }
    }
}

/*
**        process a decoded pixel (every last one thru here)
**
*/

static void putpixel(val)
unsigned char val;
{
    static short inctable[5] = {8,8,4,2,0};   /* y incrs for interlace */
    static short starttable[5] = {0,4,2,1,0}; /* starting rows for each pass */
    short key;

    put_pixel(x,y,val);     /* local copy */
    outrow[x++] = val;      /* save the pixel and bump x coord */
    if (--rowcnt == 0)      /* detally and check for end of row */
    {
        printf("\rinput row %d ",y);
        if (kbhit())
        {
            key = getch();
            if (key == 27)
            {
                sysabort("aborted by Esc");
            }
        }
        x = 0;                  /* reset x index */
        rowcnt = iw;            /* and pixel counter */
        if (interlaced)         /* generate new y interlaced */
        {
            y += inctable[pass];    /* incr y */
            if (y >= ih)            /* time to start a new pass */
            {
                y = starttable[++pass];     /* new start y */
            }
        }
        else if (++y >= ih) y = 0;      /* normal y incr, wrap if too big */
    }
}


static void putx(code, psize)
int code, psize;
{
    short ccount = 0;
    unsigned char *s;

    s=ostack;
    do
    {
        *s++ = ctlast[code];
        ccount++;
        code = ctlink[code];
    } while (code != -1);
    if (psize == 1) {
        do
        {
            putpixel(*--s & 1);
            putpixel(*s >> 1);
        } while (--ccount);
    }
    else do {
        putpixel(*--s);
    } while (--ccount);
}


/*
**        process an image block
**
*/

static void ximage(codestart,pixelsize)
int codestart,pixelsize;
{
    short reqct, code, cc, eoi, oldcode;

    cc = 1 << codestart;        /* compute clear code */
    eoi = cc + 1;               /*  and end of info code */
    reqct = codestart + 1;      /* */
    init_codetable(cc);         /* initialise the code tables */

    /* set initial conditions for the image block */

    oldcode = -1;
    done = 0;
    pass = 0;
    rowcnt = iw;
    x = 0;
    y = 0;

    do 
    {
        code = getcode(reqct);
        if (code == cc)         /* 'tis a clear code */
        {
            init_codetable(cc);     /* new code table */
            reqct = codestart + 1;  /* new ?? */
            oldcode = -1;
        }
        else if (code == eoi)   /* end of info code */
        {
            flushin();          /* flush any remaining blocks */
            done = -1;          /* and quit */
        }
        else    /* process normal code value */
        {
            if (ctlink[code] != -2)     /* a new occurrence */
            {
                if (oldcode != -1)
                {
                    insert_code(code, oldcode, &reqct);
                }
            }
            else
            {
                insert_code(oldcode, oldcode, &reqct);
            }
            putx(code, pixelsize);      /* output the code */
            oldcode = code;
        }
    } while (!done);
}

/*
**        fill screen with background color
**
*/

static void bgfill(height, width, bg)
int height, width;
unsigned char bg;
{
}



static void getcolors(psize)
int psize;
{
    short cindex, cmax;
    short r, g, b;
    unsigned char *p;

    p = palette;
    cmax = 1 << psize;
    for (cindex = 0; cindex < cmax; cindex++)
    {
        
        r = nextbyte();              /* red */
        g = nextbyte();              /* green */
        b = nextbyte();              /* blue */
#if 0
        r = (r >> 2) & 0x3f;        /* convert to 6 bits */
        g = (g >> 2) & 0x3f;
        b = (b >> 2) & 0x3f;
#endif
        *p++ = r;                   /* red */
        *p++ = g;                   /* green */
        *p++ = b;                   /* blue */
    }
}



static short gword()
{
    return(nextbyte() | (nextbyte() << 8));
}




/*
**         convert palette to grey scale
**
**      This routine is not used for the Phaser IIsd output but
**      would be needed if a grey-scale only printer is used ...
**
*/

void set_grey_scale(
int nz)                 /* number of intensities to map */
{
    short i, z;
    double red, green, blue, grey;
    unsigned char *p;
    double scale;

    scale = nz;

    p = palette;
    for (i = 0; i < 256; i++)       /* for each palette entry */
    {
        red = ((double) p[0]) / 256.0;
        green = ((double) p[1]) / 256.0;
        blue = ((double) p[2]) / 256.0;
        grey = (red * 0.30) + (green * 0.59) + (blue * 0.11);

        /* 0 <= grey < 1.0 */

        grey *= scale;      /* 0 <= grey < (10.0 or 26.0) */
        z = grey;           /* 0 <= grey < nz */
        hpalette[i] = z;    /* save the halftone value */
        p += 3;
    }
}


/*
**         allocate image memory
**
*/

static allocate_image()
{
    extern void *malloc();
    byte *b;
    short col;

    for (col = 0; col < ncols; col++)
    {
        b = (byte *) malloc(nrows);
        if (b == NULL) sysabort("image allocation failed");
        columns[col] = b;
        memset(b,255,nrows);      /* clear it to white */
    }
}

void close_gif(void)        /* release the space */
{
    short col;

    for (col = 0; col < ncols; col++)
    {
        if (columns[col]) free(columns[col]);        
    }
}

/*
**         set a pixel into the local array
**
*/

static put_pixel(x,y,z)
int x, y, z;
{
    byte *p;

    if (x >= ncols) return;     /* clip to physical buffer */
    if (y >= nrows) return;

    p = columns[x];     /* start of column x */
    p[y] = z;           /* set the pig ... */
}


/*
**         get a pixel from the local array
**
*/

int read_gif_pixel(
int x, int y)
{
    byte *p;
    byte z;

    if (x >= ncols) return (0);     /* clip to physical buffer */
    if (y >= nrows) return (0);

    p = columns[x];         /* start of column x */
    z = p[y] & 0xff;        /* get the pig ... */
    return(hpalette[z]);    /* return scaled value */
}


void zrgif(
int *xsize,             /* returned x size */
int *ysize,             /* returned y size */
int nz,                 /* number of grey scales needed */
char *zrpath)
{
    char arg2[32];
    short    n;
    char *mm, *matrix, *ctxt;
    short    alldone, height, width, gpix, pixelsize, csize, blen;
    unsigned char sflags, bg, work, iflags;
    char namework[64];

    *xsize = 0;
    *ysize = 0;
    strcpy(namework,zrpath);
    inf = fopen(namework, "rb");
    if (inf == NULL) {
        strcat(namework,".gif");
        inf = fopen(namework, "rb");
        if (inf == NULL)
        {
            sysabort("cannot open %s",namework);
        }
    }
    if (nextbyte() != 'G')
    {
        sysabort("%s is not a GIF file",namework);
    }
    for (n = 1; n <= 5; n++) nextbyte();    /* flush the gif name */

    width = gword();
    height = gword();

    sflags = nextbyte();
    gpix = (sflags & 7) + 1;

    bg = nextbyte();


    if (nextbyte()) {
#if 0       /* this was turned off for FRACTINT compatibility */
        printf("Invalid format");
        exit(1);
#endif
    }
    if (width > MAXNCOLS)
    {
        sysabort("image wider than %d columns",MAXNCOLS);
    }

    nrows = height;
    ncols = width;
    *xsize = ncols;
    *ysize = nrows;

    allocate_image();

    initscreen(height, width);

    if (sflags & 0x80) getcolors(gpix);

    bgfill(height, width, bg);

    alldone = 0;


    while (!alldone && ((work = nextbyte()) != EOF)) /* process images ... */
    {
        if (work == ',')        /* image separator */
        {
            ix = gword() + xadj;        /* center the image */
            iy = gword() + yadj;        /* ix,iy is 0,0 of screen */
            iw = gword();               /* width of image */
            ih = gword();               /* height of image */

            iflags = nextbyte();         /* grab the flag byte */
            interlaced = iflags & 0x40; /* save the interlaced bit */
            pixelsize = gpix;           /* and # bits/pixel (global) */

            if (iflags & 0x80)      /* process a local palette */
            {
                /* this code will not revert to the global palette */
                     /* the latest will always be used */

                pixelsize = (iflags & 7) + 1;   /* use local pixel size */
                getcolors(pixelsize);           /* get the new palette */
            }

            /* start processing lzw stuff ... */

            csize = nextbyte();      /* fetch initial code size */
            bufct = 0;              /*  and reset the buffer processor */

            ximage(csize,pixelsize);    /* process the image block */
        }
        else if (work == '!')           /* flush an extension block */
        {
            nextbyte();                  /* flush the function code */
            for (;;)
            {
                blen = nextbyte();    /* until a zero length block */
                if (blen == 0) break;
                for (n = 0; n < blen; n++) nextbyte();   /* flush block */
            }
        }
        else if (work == ';')    /* process a terminator */
        {
            alldone = -1;
        }
        else            /* etwas ist los! */
        {
            closescreen();
            printf("Invalid Format");
            exit(1);
        }
    }
#if 1
    set_grey_scale(nz);
#endif
    fclose(inf);
    closescreen();
}

